import asyncio
import sys

from .config import load_config
from .logging import logging_context
from .runner import TaskRunner


def main() -> None:
    config = load_config()
    with logging_context(config.log_levels):
        success = asyncio.run(TaskRunner(config.tasks).run_tasks())
    sys.exit(not success)



import logging

from .formatting import bright, dim, status
from .logging import ASYNC_LOG, task_log
from .utils import concurrent_call, exec_command, timed_awaitable

log = ASYNC_LOG


class TaskRunner:
    def __init__(self, tasks: dict[str, str]) -> None:
        self.tasks = tasks
        self.max_name_length = max(len(n) for n in tasks) if tasks else 0

    async def run_tasks(self) -> bool:
        if not self.tasks:
            log.warning('Nothing to do. Getting bored...\n')
            return True
        tasks = self.tasks.items()
        summary = [f'• {name.ljust(self.max_name_length)}  {dim(task)}' for name, task in tasks]
        for line in (bright('To do:'), *summary, '', bright('Results:')):
            log.info(line)
        results = await concurrent_call(self._run_task, tasks)
        failed_tasks = [task for task, ok in results if not ok]
        if failed_tasks:
            log.error('')
            log.error(f'{bright("Failed tasks:")} {failed_tasks}')
        return not failed_tasks

    async def _run_task(self, name: str, task: str) -> tuple[str, bool]:
        (success, (out, err)), duration = await timed_awaitable(exec_command(task))
        log.info(f'{status(success)} {name.ljust(self.max_name_length)}  {dim(duration)}')
        for level, stream in ((logging.INFO, out), (logging.ERROR, err)):
            if stream:
                task_log(success).log(level, stream.decode())
        return name, success



import argparse
import tomllib
from collections.abc import Sequence
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

from chili import decode

from .logging import LogLevel, LogLevels


@dataclass
class Config:
    tasks: dict[str, str] = field(default_factory=dict)
    log_levels: LogLevels = field(default_factory=LogLevels)

    @classmethod
    def decode(cls, value: dict) -> 'Config':
        return decode(value, Config, decoders={LogLevel: LogLevel})


class FatalError(SystemExit):
    def __init__(self, *args):
        super().__init__(f'💀 {" ".join(str(arg) for arg in args)}')


class ConfigError(FatalError):
    def __init__(self, config_source: str = None, *args):
        message = 'Could not load config'
        if config_source:
            message += f' from {config_source}'
            if args:
                message += ': ' + ' '.join(str(a) for a in args)
        super().__init__(message)


class ParseDict(argparse.Action):
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str = None,
    ) -> None:
        value_seq = [values] if isinstance(values, str) else [str(v) for v in values or []]
        try:
            pairs = (item.split('=', 1) for item in value_seq)
            new_pairs = {key.strip(): value for key, value in pairs}
        except ValueError:
            parser.error(
                f'argument {option_string}: not matching pattern key1=val1 [key2=val2 ...]',
            )
        else:
            setattr(namespace, self.dest, (getattr(namespace, self.dest) or {}) | new_pairs)


def config_from_args(_config_source: str) -> Config | None:
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument(
        '-t',
        '--tasks',
        dest='tasks',
        nargs='+',
        metavar='NAME=COMMAND',
        action=ParseDict,
        default={},
    )
    arg_parser.add_argument(
        '-l',
        '--log-levels',
        dest='log_levels',
        nargs='+',
        metavar='NAME=COMMAND',
        action=ParseDict,
        default={},
    )
    config_dict = arg_parser.parse_args().__dict__
    return Config.decode(config_dict) if any(config_dict.values()) else None


def config_from_pyproject(config_source: str) -> Config | None:
    try:
        with Path(config_source).open('rb') as f:
            config_dict = tomllib.load(f).get('tool', {}).get('powerchord', {})
    except OSError:
        return None
    return Config.decode(config_dict)


CONFIG_LOADERS = {
    'command line': config_from_args,
    'pyproject.toml': config_from_pyproject,
}


def load_config() -> Config:
    for name, loader in CONFIG_LOADERS.items():
        try:
            config = loader(name)
        except ValueError as exc:
            raise ConfigError(name, *exc.args) from exc
        if config:
            return config
    raise ConfigError



import asyncio
from collections.abc import AsyncIterator, Awaitable, Callable, Coroutine, Iterable
from subprocess import PIPE
from time import perf_counter_ns
from typing import Any, ParamSpec, TypeVar

P = ParamSpec('P')
T = TypeVar('T')


async def exec_command(command: str) -> tuple[bool, tuple[bytes, bytes]]:
    proc = await asyncio.create_subprocess_shell(command, stdout=PIPE, stderr=PIPE)
    output_streams = await proc.communicate()
    return proc.returncode == 0, output_streams


async def concurrent_iter(
    coros: Iterable[Coroutine[Any, Any, T]],
) -> AsyncIterator[T]:
    tasks: list[asyncio.Task[T]] = [asyncio.create_task(coro) for coro in coros]
    for task in tasks:
        yield await task


async def concurrent_list(
    coros: Iterable[Coroutine[Any, Any, T]],
) -> list[T]:
    return [item async for item in concurrent_iter(coros)]


async def concurrent_call(
    async_func: Callable[P, Coroutine[Any, Any, T]],
    args_list: Iterable[P.args],
) -> list[T]:
    return await concurrent_list(async_func(*args) for args in args_list)


def human_readable_duration(nanoseconds: int) -> str:
    minutes = int(nanoseconds // 60_000_000_000)
    nanoseconds %= 60_000_000_000
    seconds = int(nanoseconds // 1_000_000_000)
    nanoseconds %= 1_000_000_000
    milliseconds = int(nanoseconds // 1_000_000)
    nanoseconds %= 1_000_000
    microseconds = int(nanoseconds // 1_000)
    nanoseconds %= 1_000
    if minutes:
        return f'{minutes:d}:{seconds:02d}.{milliseconds:03d} minutes'
    if seconds:
        return f'{seconds:d}.{milliseconds:03d} seconds'
    if milliseconds:
        return f'{milliseconds:d}.{microseconds:03d} ms'
    return f'{microseconds:d}.{nanoseconds:03d} µs'


def timed(
    func: Callable[[], T],
    formatter: Callable[[int], str] = None,
) -> tuple[T, str]:
    start = perf_counter_ns()
    return func(), (formatter or human_readable_duration)(perf_counter_ns() - start)


def timed_awaitable(
    awaitable: Awaitable[T],
    formatter: Callable[[int], str] = None,
) -> Awaitable[tuple[T, str]]:
    async def wrapper() -> tuple[T, str]:
        start = perf_counter_ns()
        return await awaitable, (formatter or human_readable_duration)(perf_counter_ns() - start)
    return wrapper()
